# -*- coding: utf-8 -*-
"""
This module defines functions for parsing `STAR`_ (Self-defining Text Archiving and Retrieval) files.
This includes metadata files from cryo-EM image analysis programs including RELION and XMIPP as well as
the Crystallographic Information File (CIF) format used for much of the PDB.

.. _STAR: https://www2.mrc-lmb.cam.ac.uk/relion/index.php/Conventions_%26_File_formats#The_STAR_format
"""

from collections import OrderedDict
import os.path
from numbers import Integral
import numpy as np
import sys

from prody.utilities import openFile, split, pystr, isListLike
from prody import LOGGER

from .emdfile import parseEMD

__all__ = ['parseSTAR', 'writeSTAR', 'parseImagesFromSTAR',
           'StarDict', 'StarDataBlock', 'StarLoop', 
           'parseSTARSection']


class StarDict:
    def __init__(self, parsingDict, prog, title='unnamed', indices=None):
        self._title = title
        self._dict = parsingDict
        self._prog = prog
        self._indices = indices

        if indices is None:
            self.dataBlocks = [StarDataBlock(self, key)
                               for key in self._dict.keys()]
        else:
            self.dataBlocks = []
            for idx in indices:
                if isListLike(idx):
                    self.dataBlocks.append(StarDataBlock(self, idx[0], idx[1]))
                else:
                    self.dataBlocks.append(StarDataBlock(self, idx))

            self._dict = OrderedDict()
            for i, idx in enumerate(indices):
                self._dict[idx[0]] = self.dataBlocks[i]._dict

        self._n_data_blocks = len(self.dataBlocks)

    def numDataBlocks(self):
        return self._n_data_blocks
    
    def __getitem__(self, key):
        try:
            return np.array(self.dataBlocks)[key]
        except:
            try:
                key = np.where(np.array(list(self._dict.keys())) == key)[0][0]
                return self.dataBlocks[key]
            except:
                raise ValueError('The key for getting items should '
                                 'be the names or numbers of data blocks')

    def getTitle(self):
        return self._title

    def setTitle(self, value):
        self._title = value

    def getDict(self):
        return self._dict

    def __repr__(self):
        if self._n_data_blocks == 1:
            return '<StarDict: {0} (1 data block)>'.format(self._title)
        return '<StarDict: {0} ({1} data blocks)>'.format(self._title, self.numDataBlocks())

    def __iter__(self):
        """Yield StarDataBlock instances."""
        for key in list(self._dict.keys()):
            yield StarDataBlock(self, key)

    def pop(self, index):
        """Pop dataBlock with the given index from the list of dataBlocks in StarDict"""
        self.dataBlocks.pop(index)
        self._n_data_blocks -= 1

    def search(self, substr, return_indices=False):
        indices = []
        for data_block in self:
            if data_block.search(substr).numEntries() != 0 or data_block.search(substr).numLoops() != 0:
                indices.append((data_block._title,
                                data_block.search(substr, return_indices=True)[0]))

        if return_indices:
            return indices, StarDict(self._dict, self._prog, self._title, indices=indices)

        return StarDict(self._dict, self._prog, self._title, indices=indices)

    def printData(self):
        for data_block in self:
            sys.stdout.write('data_' + data_block._title + '\n')
            data_block.printData()
            sys.stdout.write('\n')


class StarDataBlock:
    def __init__(self, starDict, key, indices=None):
        self._title = key
        self._prog = starDict._prog
        self._starDict = starDict

        if indices is None:
            try:
                self._dict = starDict._dict[key]
            except:
                self._dict = list(starDict._dict)[key]

            keys = list(self._dict.keys())
        else:
            keys = [idx[0] for idx in indices]
            self._dict = OrderedDict()
            self._dict['data'] = OrderedDict()
            self._dict['fields'] = OrderedDict()
            for idx in indices:
                if idx[0] == 'data':
                    self._dict[idx[0]][idx[1]] = starDict._dict[self._title][idx[0]][idx[1]]
                    if not 'fields' in keys:
                        for k, v in self._starDict._dict[self._title]['fields'].items():
                            if v == idx[1]:
                                self._dict['fields'][k] = v
                else:
                    self._dict[idx[0]] = OrderedDict()
                    self._dict[idx[0]]['fields'] = starDict._dict[self._title][idx[0]]['fields']
                    self._dict[idx[0]]['data'] = OrderedDict()
                    for id1 in idx[1]:
                        self._dict[idx[0]]['data'][id1] = starDict._dict[self._title][idx[0]]['data'][id1]

        if set(keys) == set(['data', 'fields']):
            self.loops = []
            self._n_loops = 0

            self.data = np.array(list(self._dict['data'].values()))
            self.fields = np.array(list(self._dict['fields'].values()))

            if not isListLike(self.data):
                self.data = [self.data]

            if not isListLike(self.fields):
                self.fields = [self.fields]

            self._n_entries = len(self.data)
            self._n_fields = len(self.fields)

        elif 'data' in keys and 'fields' in keys:
            if indices is not None:
                self.loops = [StarLoop(self, key, idx)
                              for (key, idx) in indices
                              if key not in ['data', 'fields']]
            else:
                self.loops = [StarLoop(self, key) for key in keys
                              if key not in ['data', 'fields']]

            self.data = np.array(list(self._dict['data'].values()))
            self.fields = np.array(list(self._dict['fields'].values()))

            if not isListLike(self.data):
                self.data = [self.data]

            if not isListLike(self.fields):
                self.fields = [self.fields]

            self._n_entries = len(self.data)
            self._n_fields = len(self.fields)
            self._n_loops = len(self.loops)

        elif 'data' in keys:
            if indices is not None:
                self.loops = [StarLoop(self, key, idx)
                              for (key, idx) in indices
                              if key != 'data']
            else:
                self.loops = [StarLoop(self, key)
                              for key in keys
                              if key != 'data']

            self.data = np.array(list(self._dict['data'].values()))
            self.fields = np.array(list(self._dict['fields'].values()))

            if not isListLike(self.data):
                self.data = [self.data]

            if not isListLike(self.fields):
                self.fields = [self.fields]

            self._n_loops = len(self.loops)
            self._n_entries = len(self.data)
            self._n_fields = 0

        elif 'fields' in keys:
            if indices is not None:
                self.loops = [StarLoop(self, key, idx)
                              for (key, idx) in indices
                              if key != 'fields']
            else:
                self.loops = [StarLoop(self, key)
                              for key in keys
                              if key != 'fields']

            self.data = np.array(list(self._dict['data'].values()))
            self.fields = np.array(list(self._dict['fields'].values()))

            if not isListLike(self.data):
                self.data = [self.data]

            if not isListLike(self.fields):
                self.fields = [self.fields]

            self._n_loops = len(self.loops)
            self._n_entries = len(self.data)
            self._n_fields = 0

        else:
            if indices is not None:
                self.loops = [StarLoop(self, key, idx)
                              for (key, idx) in indices]
            else:
                self.loops = [StarLoop(self, key) for key in keys]

            self._n_loops = len(self.loops)
            self._n_entries = 0
            self._n_fields = 0

    def numLoops(self):
        return self._n_loops
    
    def numEntries(self):
        return self._n_entries
    
    def getLoop(self, index):
        try:
            return self.loops[index]
        except:
            raise ValueError('There is no loop with that index')

    def getTitle(self):
        return self._title

    def setTitle(self, title):
        self._title = title

    def getDict(self):
        return self._dict

    def __getitem__(self, key):
        if key == 'data':
            try:
                return self._dict[key]
            except:
                raise ValueError('This StarDataBlock has no non-loop data')
        if key == 'fields':
            try:
                return self._dict[key]
            except:
                raise ValueError(
                    'This StarDataBlock has no non-loop data fields')
        else:
            try:
                return StarLoop(self, key)
            except:
                try:
                    return self.loops[key]
                except:
                    raise ValueError('The key for getting items should be data, fields, '
                                     'or the name or number of a loop')

    def __repr__(self):
        if self.numLoops() == 0:
            if self.numEntries() == 0:
                return '<StarDataBlock: {0} (no entries)>'.format(self._title)
            if self.numEntries() == 1:
                return '<StarDataBlock: {0} (1 entry)>'.format(self._title)
            else:
                return '<StarDataBlock: {0} ({1} entries)>'.format(self._title, self.numEntries())
        elif self.numEntries() == 0:
            if self.numLoops() == 1:
                return '<StarDataBlock: {0} (1 loop containing ' \
                    '{1} columns and {2} rows)>'.format(self._title,
                                                        self.loops[0].numFields(), self.loops[0].numRows())
            return '<StarDataBlock: {0} ({1} loops)>'.format(self._title, self.numLoops())
        else:
            if self.numLoops() == 1:
                if self.numEntries() == 1:
                    return '<StarDataBlock: {0} (1 entry and 1 loop)>'.format(self._title)
                else:
                    return '<StarDataBlock: {0} ({1} entries and 1 loop)>'.format(self._title, self.numEntries())
            else:
                if self.numEntries() == 1:
                    return '<StarDataBlock: {0} (1 entry and {1} loops)>'.format(self._title, self.numLoops())
                else:
                    return '<StarDataBlock: {0} ({1} entries and {2} loops)>'.format(self._title, self.numEntries(),
                                                                                     self.numLoops())

    def __iter__(self):
        """Yield StarLoop instances."""
        for key in list(self._dict.keys()):
            yield StarLoop(self, key)

    def pop(self, index):
        """Pop loop with the given index from the list of loops in dataBlock"""
        self.loops.pop(index)
        self._n_loops -= 1

    def search(self, substr, return_indices=False):
        indices = []
        for key, value in self._dict.items():
            if key == 'fields':
                pass
            elif key == 'data':
                for field, val in value.items():
                    if field.find(substr) != -1 or val.find(substr) != -1:
                        indices.append(('data', field))
            else:
                idx, loop = self[key].search(substr, return_indices=True)
                if loop.numRows() != 0:
                    indices.append((key, idx))

        if return_indices:
            return indices, StarDataBlock(self._starDict, self._title, indices)

        return StarDataBlock(self._starDict, self._title, indices)

    def printData(self):
        for key, value in self._dict.items():
            if key == 'fields':
                longest_len = 0
                for field in self[key].values():
                    if len(field) > longest_len:
                        longest_len = len(field)
            elif key == 'data':
                longest_len = 0
                for field, val in value.items():
                    sys.stdout.write(field + ' '*(longest_len - len(field)))
                    sys.stdout.write('\t' + val + '\n')
                sys.stdout.write('\n')
            else:
                sys.stdout.write('_loop\n')
                self[key].printData()
                sys.stdout.write('\n')


class StarLoop:
    def __init__(self, dataBlock, key, indices=None):
        self._key = key
        self._dataBlock = dataBlock

        if indices is None:
            self._dict = dataBlock._dict[self._key]
        else:
            self._dict = OrderedDict()
            self._dict['fields'] = dataBlock._dict[self._key]['fields']
            self._dict['data'] = OrderedDict()

            indices_are_keys = True
            for index in indices:
                try:
                    self._dict['data'][index] = dataBlock._dict[self._key]['data'][index]
                except:
                    indices_are_keys = False
                    break

            if not indices_are_keys:
                self._dict['data'] = OrderedDict()
                for index in indices:
                    self._dict['data'][index] = list(self._dataBlock._dict[self._key]['data'].values())[index]

        self._prog = dataBlock._prog
        self.fields = list(self._dict['fields'].values())
        self.data = list(self._dict['data'].values())
        self._n_fields = len(self.fields)
        self._n_rows = len(self.data)
        self._title = str(dataBlock._title) + ' loop ' + str(key)

    def numRows(self):
        return self._n_rows
    
    def numFields(self):
        return self._n_fields
    
    def getData(self, key):
        if key in self.fields:
            return [row[key] for row in self.data]
        else:
            raise ValueError('That field is not present in this loop')

    def getTitle(self):
        return self._title

    def setTitle(self, title):
        self._title = title

    def getDict(self):
        return self._dict

    def __getitem__(self, key):
        try:
            return np.array(self.data)[key]
        except:
            try:
                try:
                    return self._dict[key]
                except KeyError:
                    key = list(self._dict.keys()).index(key)
                    return self.data[key]
            except:
                try:
                    return self.getData(key)
                except:
                    raise ValueError('The key for getting items should be fields, data, '
                                     'or a field name or number corresponding to a '
                                     'row or column of data')

    def __repr__(self):
        if self.numFields() == 1 and self.numRows() != 1:
            return '<StarLoop: {0} (1 column and {1} rows)>'.format(self._title, self.numRows())
        elif self.numFields() != 1 and self.numRows() == 1:
            return '<StarLoop: {0} ({1} columns and 1 row)>'.format(self._title, self.numFields())
        elif self.numFields() == 1 and self.numRows() == 1:
            return '<StarLoop: {0} (1 column and 1 row)>'.format(self._title)
        else:
            return '<StarLoop: {0} ({1} columns and {2} rows)>'.format(self._title, self.numFields(), self.numRows())

    def search(self, substr, return_indices=False):
        indices = []
        for j, row in self._dict['data'].items():
            found_it = False
            for entry in row.items():
                field, value = entry
                if field.find(substr) != -1 or value.find(substr) != -1:
                    found_it = True
                    break
            if found_it:
                indices.append(j)

        if return_indices:
            return indices, StarLoop(self._dataBlock, self._key, indices)

        return StarLoop(self._dataBlock, self._key, indices)

    def printData(self):
        for field in self.fields:
            sys.stdout.write(field + '\n')
        for row in self.data:
            for entry in row.values():
                sys.stdout.write(entry + '\t')
            sys.stdout.write('\n')


def parseSTAR(filename, **kwargs):
    """Returns a dictionary containing data parsed from a STAR file.

    :arg filename: a filename
        The .star extension can be omitted.
    :type filename: str

    :arg start: line number for starting
        Default is **None**, meaning start at the beginning
    :type start: int, None

    :arg stop: line number for stopping
        Default is **None**, meaning don't stop.
    :type stop: int, None

    :arg shlex: whether to use shlex for splitting lines so as to preserve quoted substrings
        Default is **False**
    :type shlex: bool
    """
    if not os.path.isfile(filename) and not os.path.isfile(filename + '.star'):
        raise IOError('There is no file called {0}.'.format(filename))

    start = kwargs.get('start', None)
    if start is not None and not isinstance(start, Integral):
        raise TypeError('start should be an integer or None')

    stop = kwargs.get('stop', None)
    if stop is not None and not isinstance(stop, Integral):
        raise TypeError('stop should be an integer or None')

    shlex = kwargs.get('shlex', False)
    if not isinstance(shlex, bool):
        raise TypeError('shlex should be a boolean')

    starfile = openFile(filename, 'r')
    lines = [pystr(line) for line in starfile.readlines()]
    starfile.close()

    parsingDict, prog = parseSTARLines(lines, **kwargs)

    return StarDict(parsingDict, prog, filename)


def parseSTARLines(lines, **kwargs):
    start = kwargs.get('start', None)
    if start is None:
        start = 0

    stop = kwargs.get('stop', None)
    if stop is None:
        stop = len(lines)

    prog = kwargs.get('prog', None)
    shlex = kwargs.get('shlex', False)

    finalDictionary = OrderedDict()
    currentLoop = -1
    block_fieldCounter = 0
    loop_fieldCounter = 0
    active_fieldCounter = 0
    dataItemsCounter = 0
    lineNumber = 0
    inLoop = False
    inShortBlock = False
    for line in lines[start:stop]:
        if line.startswith('data_'):
            currentDataBlock = line[5:].strip()
            finalDictionary[currentDataBlock] = OrderedDict()
            currentLoop = -1
            inLoop = False
            inShortBlock = False
            startingBlock = True
            block_fieldCounter = 0

        elif line.startswith('loop_'):
            currentLoop += 1
            inLoop = True
            inShortBlock = False
            finalDictionary[currentDataBlock][currentLoop] = OrderedDict()
            finalDictionary[currentDataBlock][currentLoop]['fields'] = OrderedDict()
            finalDictionary[currentDataBlock][currentLoop]['data'] = OrderedDict()
            loop_fieldCounter = 0

        elif line.startswith('_') or line.startswith(' _'):
            # This marks a field identifier
            currentField = split(line.strip(), shlex=shlex)[0]

            if inLoop:
                # We expect to only have the field identifier and no data until after

                if len(split(line.strip(), shlex=shlex)) == 1 or len(split(line.strip(), shlex=shlex)) == 2:
                    # This is what we expect for a data loop
                    finalDictionary[currentDataBlock][currentLoop]['fields'][loop_fieldCounter] = currentField
                    dataItemsCounter = 0
                    loop_fieldCounter += 1

                else:
                    # This is contrary to that so we leave the loop
                    inLoop = False

                    # We populate fields and data together, continuing the regular data block
                    finalDictionary[currentDataBlock]['fields'][block_fieldCounter] = currentField
                    finalDictionary[currentDataBlock]['data'][currentField] = split(line.strip(),
                                                                                    shlex=shlex)[1]
                    block_fieldCounter += 1

            else:
                # Outside a loop, populate fields and data together in a regular data block
                if startingBlock:
                    # Initialise the data block first
                    finalDictionary[currentDataBlock]['fields'] = OrderedDict()
                    finalDictionary[currentDataBlock]['data'] = OrderedDict()
                    startingBlock = False
                    block_fieldCounter = 0

                finalDictionary[currentDataBlock]['fields'][block_fieldCounter] = currentField

                if len(split(line.strip(), shlex=shlex)) > 1:
                    # This is the usual behaviour so we can fill in the data from the rest of the line
                    finalDictionary[currentDataBlock]['data'][currentField] = split(line.strip(),
                                                                                    shlex=shlex)[1]
                else:
                    # In this case, we will look for the data in a short block over the following line(s).
                    # If a single field takes multiple lines, these lines start and end with a semi-colon.
                    # We'll handle that in the data section.
                    finalDictionary[currentDataBlock]['data'][currentField] = ''
                    inShortBlock = True
                    startingShortBlock = True

                block_fieldCounter += 1

        elif line.strip() == '#':
            inLoop = False
            inShortBlock = False

        elif line.strip() == '':
            pass

        elif inLoop:
            # Here we handle the data part of the loop.
            # Data outside a loop is handled in line with the fields above or in shortDataBlocks below.
            if not inShortBlock and len(split(line, shlex=shlex)) == loop_fieldCounter:
                # This is the usual case where each entry in the line corresponds to a field
                finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter] = OrderedDict()
                active_fieldCounter = 0
                for fieldEntry in split(line.strip(), shlex=shlex):
                    currentField = finalDictionary[currentDataBlock][currentLoop]['fields'][active_fieldCounter]
                    finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter][currentField] = fieldEntry
                    active_fieldCounter += 1
                dataItemsCounter += 1
            else:
                # The data is now being broken across lines.
                if not inShortBlock:
                    inShortBlock = True
                    finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter] = OrderedDict()
                    active_fieldCounter = 0
                    if not line.startswith(';'):
                        # Then we haven't got a split field and can treat fields as normal
                        inSplitField = False
                        for fieldEntry in split(line.strip(), shlex=shlex):
                            currentField = finalDictionary[currentDataBlock][currentLoop]['fields'][active_fieldCounter]
                            finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter][currentField] = fieldEntry
                            active_fieldCounter += 1
                    else:
                        # We have a single field split over many lines
                        inSplitField = True
                        currentField = finalDictionary[currentDataBlock][currentLoop]['fields'][active_fieldCounter]
                        finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter][currentField] = line.strip() + ' '
                else:
                    if not inSplitField:
                        # check if we are entering one
                        if line.startswith(';'):
                            inSplitField = True
                            currentField = finalDictionary[currentDataBlock][currentLoop]['fields'][active_fieldCounter]
                            finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter][currentField] = line.strip() + ' '
                        else:
                            # continue as normal
                            for fieldEntry in split(line.strip(), shlex=shlex):
                                currentField = finalDictionary[currentDataBlock][currentLoop]['fields'][active_fieldCounter]
                                finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter][currentField] = fieldEntry
                                active_fieldCounter += 1
                    else:
                        finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter][currentField] += line.strip()
                        if line.strip() == ';':
                            # This marks the end of the split field
                            inSplitField = False
                            active_fieldCounter += 1
                        else:
                            # Prepare for the next line
                            finalDictionary[currentDataBlock][currentLoop]['data'][dataItemsCounter][currentField] += ' '

                    if active_fieldCounter == loop_fieldCounter:
                        inShortBlock = False
                        dataItemsCounter += 1

        elif inShortBlock:
            # We can now append the data in the lines here.
            finalDictionary[currentDataBlock]['data'][currentField] += line.strip()
            if startingShortBlock:
                startingShortBlock = False
                if not line.startswith(';'):
                    # We only expect one line if there's no semi-colon
                    inShortBlock = False
                else:
                    # Prepare for the next line
                    finalDictionary[currentDataBlock]['data'][currentField] += ' '
            else:
                if line.strip() == ';':
                    # This marks the end of the field so we've filled it
                    inShortBlock = False
                else:
                    # Prepare for the next line
                    finalDictionary[currentDataBlock]['data'][currentField] += ' '

        elif line.startswith('#'):
            if line.startswith('# XMIPP'):
                prog = 'XMIPP'

        else:
            raise TypeError('This file does not conform to the STAR file format. '
                            'There is a problem with line {0}:\n {1}'.format(lineNumber, line))

        lineNumber += 1

    return finalDictionary, prog


def writeSTAR(filename, starDict, **kwargs):
    """Writes a STAR file from a dictionary containing data
    such as that parsed from a Relion STAR file.

    :arg filename: a filename
        The .star extension can be omitted.
    :type filename: str

    :arg starDict: a dictionary in STAR format
        This should have nested entries starting with data blocks then loops/tables then
        field names and finally data.
    :type starDict: dict

    kwargs can be given including the program style to follow (*prog*)
    """
    prog=kwargs.get('prog', 'XMIPP')

    star = open(filename, 'w')

    for dataBlockKey in starDict:
        star.write('\ndata_' + dataBlockKey + '\n')
        for loopNumber in starDict[dataBlockKey]:
            star.write('\nloop_\n')
            for fieldNumber in starDict[dataBlockKey][loopNumber]['fields']:
                if prog == 'XMIPP':
                    star.write(' ')
                star.write(starDict[dataBlockKey][loopNumber]['fields'][fieldNumber] + '\n')
            for dataItemNumber in starDict[dataBlockKey][loopNumber]['data']:
                if prog == 'XMIPP':
                    star.write('\t')
                for fieldNumber in starDict[dataBlockKey][loopNumber]['fields']:
                    currentField = starDict[dataBlockKey][loopNumber]['fields'][fieldNumber]
                    star.write(starDict[dataBlockKey][loopNumber]['data'][dataItemNumber][currentField] + '\t')
                star.write('\n')

    star.close()
    return


def parseImagesFromSTAR(particlesSTAR, **kwargs):
    """
    Parses particle images using data from a STAR file 
    containing information about them.

    :arg particlesSTAR: a filename for a STAR file.
    :type particlesSTAR: str

    :arg block_indices: indices for data blocks containing rows 
        corresponding to images of interest
        The indexing scheme is similar to that for numpy arrays.
        Default behavior is use all data blocks about images
    :type block_indices: list, :class:`~numpy.ndarray`

    :arg row_indices: indices for rows corresponding to images of interest
        The indexing scheme is similar to that for numpy arrays. 
        row_indices should be a 1D or 2D array-like.
        2D row_indices should contain an entry for each relevant loop. 
        If a 1D array-like is given the same row indices 
        will be applied to all loops.
        Default behavior is to use all rows about images
    :type row_indices: list, :class:`~numpy.ndarray`

    :arg particle_indices: indices for particles regardless of STAR structure
        default is take all particles
        Please note: this acts after block_indices and row_indices
    :type particle_indices: list, :class:`~numpy.ndarray`

    :arg saveImageArrays: whether to save the numpy array for each image to file
        default is False
    :type saveImageArrays: bool

    :arg saveDirectory: directory where numpy image arrays are saved
        default is **None**, which means save to the current working directory
    :type saveDirectory: str

    :arg rotateImages: whether to apply in plane translations and rotations using 
        provided psi and origin data, default is True
    :type rotateImages: bool 
    
    """

    try:
        from skimage.transform import rotate
    except ImportError:
        raise ImportError('This function requires scikit-image.')

    block_indices = kwargs.get('block_indices', None)
    # No loop_indices because data blocks about particle images contain 1 loop
    row_indices = kwargs.get('row_indices', None)
    particle_indices = kwargs.get('particle_indices', None)

    saveImageArrays = kwargs.get('saveImageArrays', False)
    saveDirectory = kwargs.get('saveDirectory', None)
    rotateImages = kwargs.get('rotateImages', True)

    try:
        particlesSTAR = parseSTAR(particlesSTAR)
    except:
        raise ValueError('particlesSTAR should be a filename for a STAR file')

    # Check dimensions/contents of particlesSTAR and generate full indices
    dataBlocks = []
    loops = []
    maxLoops = 0
    maxRows = 0
    dataBlock_goodness = []
    for dataBlock in particlesSTAR:

        foundImageField = False
        for loop in dataBlock:
            if ('_image' in loop.fields) or ('_rlnImageName' in loop.fields):
                foundImageField = True
                loops.append(loop)
                if loop.numRows() > maxRows:
                    maxRows = loop.numRows()
            else:
                dataBlock.pop(int(loop.getTitle().split(' ')[-1]))

        if dataBlock.numLoops() > maxLoops:
            maxLoops = dataBlock.numLoops()

        if foundImageField:
            dataBlocks.append(dataBlock)
            dataBlock_goodness.append(True)
        else:
            dataBlock_goodness.append(False)

    indices = np.zeros((len(dataBlocks), maxLoops, maxRows, 3), dtype=int)
    i = -1
    for n, dataBlock in enumerate(particlesSTAR):
        if dataBlock_goodness[n]:
            i += 1
            for j, loop in enumerate(dataBlock):
                for k in range(maxRows):
                    if k < loop.n_rows:
                        indices[i, j, k] = np.array([n, j, k])
                    else:
                        indices[i, j, k] = np.array([0, 0, 0])

    dataBlocks = np.array(dataBlocks)
    loops = np.array(loops)

    # Convert keyword indices to valid indices if possible
    if block_indices is not None:
        if np.array_equal(dataBlocks, np.array([])):
            raise TypeError(
                'particlesSTAR must have data blocks to use block_indices')

        try:
            block_indices = np.array(block_indices)
        except:
            raise TypeError('block_indices should be array-like')

        if block_indices.ndim != 1:
            raise ValueError(
                'block_indices should be a 1-dimensional array-like')

        for i, index in enumerate(list(reversed(block_indices))):
            try:
                block = particlesSTAR[index]
                if not isinstance(block, StarDataBlock):
                    LOGGER.warn('There is no block corresponding to block_index {0}. '
                                'This index has been removed.'.format(block_indices.shape[0]-i-1))
                    block_indices = np.delete(block_indices, i, 0)
            except:
                LOGGER.warn('There is no block corresponding to block_index {0}. '
                            'This index has been removed.'.format(block_indices.shape[0]-i-1))
                block_indices = np.delete(block_indices, i, 0)

        if not np.array_equal(block_indices, np.array([])):
            indices = np.concatenate(([indices[np.where(indices[:, 0, 0, 0] == item)]
                                       for item in block_indices]), axis=0)
        else:
            LOGGER.warn('None of the block_indices corresponded to dataBlocks. '
                        'Default block indices corresponding to all dataBlocks '
                        'will be used instead.')

    dataBlocks = particlesSTAR[block_indices]

    if row_indices is not None:
        try:
            row_indices = np.array(row_indices)
        except:
            raise TypeError('row_indices should be array-like')

        if row_indices.ndim == 1:
            if isinstance(row_indices[0], int):
                # row_indices provided was truly 1D so
                # we will use same row indices for all data blocks
                # and warn the user we are doing so
                if len(dataBlocks) != 1:
                    LOGGER.warn('row_indices is 1D but there are multiple data blocks '
                                'so the same row indices will be used for each')

                row_indices = np.array(
                    [row_indices for i in range(len(dataBlocks))])
                # This also works if len(dataBlocks) == 1

            elif isinstance(row_indices[0], (list, tuple)):
                # A list-like of list-likes of different sizes was provided
                # We turn it into a proper 2D array by filling the short
                # list likes with zeros

                if len(row_indices) != len(dataBlocks):
                    raise ValueError('There should be an entry in row indices for '
                                     'each data block')

                max_len = 0
                for entry in row_indices:
                    if not np.isscalar(entry):
                        if len(entry) > max_len:
                            max_len = len(entry)

                row_indices_list_entries = []
                for entry in row_indices:
                    if isinstance(entry, int):
                        list_entry = [entry]
                    else:
                        list_entry = list(entry)

                    while len(list_entry) < max_len:
                        list_entry.append(0)

                    row_indices_list_entries.append(list_entry)

                row_indices = np.array(row_indices_list_entries)

        elif row_indices.ndim == 2:
            # A list-like of list-likes of the same size was provided
            if row_indices.shape[0] != len(dataBlocks):
                if len(row_indices) == 1:
                    # we will use same row indices for all data blocks
                    # and warn the user we are doing so
                    if len(dataBlocks) != 1:
                        LOGGER.warn('row_indices has one entry but there are multiple data blocks '
                                    'so the same row indices will be used for each')

                    row_indices = np.array([row_indices[0]
                                            for i in range(len(dataBlocks))])
                    # This also works if len(dataBlocks) == 1
                else:
                    raise ValueError('There should be an entry in row indices for '
                                     'each data block')

        else:
            raise ValueError(
                'row_indices should be 1D or 2D array-like objects')

        # indices need updating
        good_indices_list = []
        for i, index_i in enumerate(indices):
            good_indices_list.append([])
            for j, index_j in enumerate(index_i):
                good_indices_list[i].append([])
                for r, index_r in enumerate(row_indices[i]):
                    for k, index_k in enumerate(index_j):
                        if k == index_r:
                            if not (r != 0 and index_r == 0):
                                good_indices_list[i][j].append(index_k)
                            else:
                                good_indices_list[i][j].append(
                                    np.array([0, 0, 0]))

        indices = np.array(good_indices_list)

    if indices is np.array([]):
        raise ValueError(
            'selection does not contain any rows with image fields')

    # Use indices to collect particle data dictionaries
    particles = []

    for i, index_i in enumerate(indices):
        for j, index_j in enumerate(index_i):
            for k, index_k in enumerate(index_j):
                if not (np.array_equal(index_k, np.array([0, 0, 0]))
                        and not (i == 0 and j == 0 and k == 0)):
                    particles.append(
                        particlesSTAR[index_k[0]][index_k[1]][index_k[2]])

    if particle_indices is None:
        particle_indices = list(range(len(particles)))

    # Parse images using particle dictionaries
    image_stacks = OrderedDict()
    images = []
    parsed_images_data = []
    stk_images = []
    if particlesSTAR._prog == 'XMIPP':
        imageFieldKey = '_image'
    else:
        imageFieldKey = '_rlnImageName'

    for i in particle_indices:
        particle = particles[i]

        try:
            image_field = particle[imageFieldKey]
            image_index = int(image_field.split('@')[0])-1
            filename = image_field.split('@')[1]
        except:
            raise ValueError('particlesSTAR does not contain data about particle image '
                             '{0} location in either RELION or XMIPP format'.format(i))

        if filename.lower().endswith('.stk'):
            stk_images.append(str(i))
            continue

        if not filename in list(image_stacks.keys()):
            image_stacks[filename] = parseEMD(filename).density

        image = image_stacks[filename][image_index]
        parsed_images_data.append(image_field)

        if saveImageArrays:
            if saveDirectory is not None:
                np.save('{0}/{1}'.format(saveDirectory, i), image)
            else:
                np.save('{1}'.format(i), image)

        if rotateImages:
            if particlesSTAR._prog == 'RELION':
                anglePsi = float(particle['_rlnAnglePsi'])
                originX = float(particle['_rlnOriginX'])
                originY = float(particle['_rlnOriginY'])
            elif particlesSTAR._prog == 'XMIPP':
                anglePsi = float(particle['_anglePsi'])
                originX = float(particle['_shiftX'])
                originY = float(particle['_shiftY'])
            images.append(rotate(image, anglePsi,
                                 center=(float(image.shape[0])-originX,
                                         float(image.shape[1])-originY)))
        else:
            images.append(image)

    if len(stk_images) > 0:
        LOGGER.warn('ProDy currently cannot parse images from XMIPP .stk files. '
                    'Please be aware that images {0} and {1} will be missing '
                    'from the final array.'.format(', '.join(stk_images[:-1]), stk_images[-1]))

    return np.array(images), parsed_images_data


def parseSTARSection(lines, key, report=True):
    """Parse a section of data from *lines* from a STAR file 
    corresponding to a *key* (part before the dot). 
    This can be a loop or data block.
    
    Returns data encapulated in a list and the associated fields.
    
    :arg report: whether to report warnings about not finding data
        default True
    :type report: bool
    """

    if not isinstance(key, str):
        raise TypeError("key should be a string")

    if not key.startswith("_"):
        key = "_" + key

    i = 0
    fields = OrderedDict()
    fieldCounter = -1
    foundBlock = False
    foundBlockData = False
    doneBlock = False
    start = 0
    stop = 0

    while not doneBlock and i < len(lines):
        line = lines[i]
        if line.split(".")[0] == key:
            fieldCounter += 1
            fields[line.split(".")[1].strip()] = fieldCounter
            if not foundBlock:
                foundBlock = True

        if foundBlock:
            if not line.startswith("#"):
                if not foundBlockData:
                    start = i
                    foundBlockData = True
            else:
                if foundBlockData:
                    doneBlock = True
                    stop = i

        i += 1

    if i < len(lines) or start > 0:
        star_dict, _ = parseSTARLines(lines[:5] + lines[start-1: stop], shlex=True)
        loop_dict = list(star_dict.values())[0]

        if lines[start - 1].strip() == "loop_":
            data = list(loop_dict[0]["data"].values())
        else:
            data = [loop_dict["data"]]
    else:
        if report:
            LOGGER.warn("Could not find {0} in lines.".format(key))
        return []

    return data



"""This module defines class for handling and storing package settings."""

import os

from .pathtools import pickle, unpickle, isWritable, USERHOME
import prody as pkg

try:
    input = raw_input
except NameError:
    pass

__all__ = ['PackageSettings', 'getPackagePath', 'setPackagePath']


class PackageSettings(object):

    """A class for managing package settings.  Settings are saved in user's
    home director.  When settings are changed by the users, the changes are
    automatically saved.  Settings are stored in a :class:`dict` instance.
    The dictionary is pickled in user's home directory for permanent storage.
    """

    def __init__(self, pkg, rcfile=None, logger=None):
        """*rcfile* is the filename for pickled settings dictionary, and by
        default is set to :file:`.pkgrc`."""

        self._package = pkg
        if rcfile is None:
            if USERHOME is not None:
                self._rcfile = os.path.join(USERHOME, '.' + pkg + 'rc')
            else:
                self._rcfile = None
        else:
            self._rcfile = rcfile

        self._logger = logger
        self._settings = {}

    def __getitem__(self, key):

        return self._settings[key]

    def __setitem__(self, key, value):

        self._settings[key] = value

    def __contains__(self, item):

        return item in self._settings

    def get(self, key, default=None):
        """Returns value corresponding to specified *key*, or *default* if *key*
        is not found."""

        return self._settings.get(key, default)

    def pop(self, key, default=None):
        """Remove specified *key* and return corresponding value.  If *key* is
        not found, *default* is returned."""

        return self._settings.pop(key, default)

    def update(self, *args, **kwargs):
        """Update settings dictionary. """

        for arg in args:
            self._settings.update(arg)
        if kwargs:
            self._settings.update(kwargs)

    def load(self):
        """Load settings by unpickling the settings dictionary."""

        if self._rcfile is not None and os.path.isfile(self._rcfile):
            try:
                settings = unpickle(self._rcfile)
            except Exception as err:
                if self._logger:
                    self._logger.warning("{0} configuration file '{1}' "
                                         "could not be loaded ({2})."
                                         .format(self._package, self._rcfile,
                                                 err))
            else:
                if isinstance(settings, dict):
                    self._settings.update(settings)

    def save(self, backup=False):
        """Save settings by pickling the settings dictionary."""

        if USERHOME is None:
            self._logger.warning("User home directory is not set, "
                                 "configuration data is not saved.")
        elif isWritable(USERHOME):
            try:
                pickle(self._settings, self._rcfile, backup=backup)
            except Exception as err:
                if self._logger:
                    self._logger.warning("{0} cannot write configuration "
                                         "file '{1}' ({2})."
                                         .format(self._package, self._rcfile,
                                                 err))
        elif self._logger:
            self._logger.warning("{0} cannot write configuration file to "
                                 "'{1}', user does not have write access."
                                 .format(self._package, USERHOME))


def setPackagePath(path):
    """Set package path."""

    if not os.path.isdir(path):
        try:
            os.mkdir(path)
        except Exception as err:
            pkg.LOGGER.warning('Failed to make folder "{0}": {1}'
                               .format(path, err.strerror))
            return False
    pkg.SETTINGS['package_path'] = path
    pkg.SETTINGS.save()
    return path


def getPackagePath():
    """Returns package path."""

    path = pkg.SETTINGS.get('package_path', None)

    update = False
    if path is None:
        pkg.LOGGER.warning('{0} package path is not yet set by the user.'
                           .format(pkg.SETTINGS._package))
        update = True
    elif not os.path.isdir(path):
        pkg.LOGGER.warning("{0} package path '{1}' does not exist."
                           .format(pkg.SETTINGS._package, path))
        update = True
    elif not os.access(path, os.W_OK):
        pkg.LOGGER.warning("User does not have write access to {0} package "
                           "path '{1}'.".format(pkg.SETTINGS._package, path))
        update = True
    if update:
        default = os.path.join(USERHOME, '.' + pkg.SETTINGS._package)
        path = input('Please specify a folder for storing {0} data '
                     '(press enter for "{1}"):'
                     .format(pkg.SETTINGS._package, default)) or default
        while not setPackagePath(path):
            path = input('Please specify a valid folder name with write '
                         'access:')
    return path



# -*- coding: utf-8 -*-
"""This module defines functions for parsing MMTF files and structure objects.

.. MMTF files: https://mmtf.rcsb.org"""

from numbers import Number
import os.path
from collections import OrderedDict

from prody.atomic import AtomGroup
from prody.atomic import flags
from prody.atomic import ATOMIC_FIELDS

from prody.utilities import openFile, isListLike, copy
from prody import LOGGER
from prody.utilities.misctools import getMasses
from .header import getHeaderDict, buildBiomolecules

import struct as st
import numpy as np

__all__ = ['parseMMTF', 'writeMMTF']

_parseMMTFdoc = """
    :arg chain: Chain identifier(s) to parse (e.g., 'A' for chain A). If not provided,
        all chains are parsed. If a PDB ID is used, chain can also be specified as part
        of the ID (e.g., '1abcA' to parse chain A).
    :type chain: str, optional

    :arg title: Title to assign to the resulting AtomGroup. If not provided, the
        title is extracted from the MMTF structure or set to the PDB ID.
    :type title: str, optional

    :arg subset: a predefined keyword to parse subset of atoms, valid keywords
        are ``'calpha'`` (``'ca'``), ``'backbone'`` (``'bb'``), or **None**
        (read all atoms), e.g. ``subset='bb'``
    :type subset: str

    :arg model: model index or None (read all models), e.g. ``model=10``
    :type model: int, list

    :arg altloc: if a location indicator is passed, such as ``'A'`` or ``'B'``,
         only indicated alternate locations will be parsed as the single
         coordinate set of the AtomGroup. If *altloc* is ``'all'`` then all
         alternate locations will be parsed and each will be appended as a
         distinct coordinate set, default is ``"A"``.  In the rare instance
         where all atoms have a location indicator specified and this does not
         match altloc, the first location indicator in the file is used.
    :type altloc: str
    """

def parseMMTF(mmtf_struc, **kwargs):
    """
    Parse an MMTF (Macromolecular Transmission Format) structure or fetch it from the PDB,
    and return an AtomGroup containing the parsed data.

    :param mmtf_struc: The MMTF structure to parse. It can be provided in one of the following ways:
        - A string representing a PDB ID (e.g., '1abc').
        - The filename of an MMTF file (ending with '.mmtf' or '.mmtf.gz').
        - An MMTF structure object (file parsed through mmtf-python).
    :type mmtf_struc: str or MMTF Structure object

    :param chain: Chain identifier(s) to parse (e.g., 'A' for chain A). If not provided,
        all chains are parsed. If a PDB ID is used, chain can also be specified as part
        of the ID (e.g., '1abcA' to parse chain A).
    :type chain: str, optional

    :return: An AtomGroup containing the parsed atomic data.
    :rtype: AtomGroup
    """
    try:
        from mmtf import fetch, parse
        import gzip, msgpack
        import mmtf
    except ImportError:
        raise ImportError("Install mmtf to read in mmtf structure objects (e.g. pip install mmtf-python)")

    chain = kwargs.pop('chain', None)
    title = kwargs.get('title', None)

    if type(mmtf_struc)==str:
        if not mmtf_struc.endswith(".mmtf") and not mmtf_struc.endswith(".mmtf.gz"): #entering just the pdb id
            if len(mmtf_struc) == 5 and mmtf_struc.isalnum():
                if chain is None:
                    chain = mmtf_struc[-1]
                    mmtf_struc = mmtf_struc[:4]
                else:
                    raise ValueError('Please provide chain as a keyword argument '
                                    'or part of the PDB ID, not both')
            else:
                chain = chain

            if len(mmtf_struc) == 4 and mmtf_struc.isalnum():
                if title is None:
                    title = mmtf_struc
                    kwargs['title'] = title
                structure = fetch(mmtf_struc)
                if structure is None:
                    raise IOError('mmtf file for {0} could not be downloaded.'
                                    .format(mmtf_struc))
                mmtf_struc = structure
            else:
                raise IOError('{0} is not a valid mmtf filename or a valid PDB '
                            'identifier.'.format(mmtf_struc))
            
        else: #entering the mmtf file name
            structure=parse(mmtf_struc)
            if structure is None:
                raise IOError('mmtf file for {0} could not be downloaded.'
                                .format(mmtf_struc))
            mmtf_struc = structure
            title = mmtf_struc.structure_id  
    elif type(mmtf_struc)==bytearray:
        data = gzip.decompress(mmtf_struc)
        unpack = msgpack.unpackb(data)
        dec = mmtf.MMTFDecoder()
        dec.decode_data(unpack)
        mmtf_struc = dec


    #if none of the above loops are entered, user should have passed a mmtf structure object
    if title is None:
        title = mmtf_struc.structure_id

    result = _parseMMTF(mmtf_struc, chain=chain, **kwargs)

    return result

parseMMTF.__doc__ += _parseMMTFdoc

def _has_biomol(hd):
    '''Return true if header actually has multiple transformations'''
    if 'biomoltrans' not in hd:
        return False
    for v in hd['biomoltrans'].values():
        if len(v) > 4:
            return True
    return False
    
def _parseMMTF(mmtf_struc, **kwargs):
    LOGGER.timeit()
    ag = AtomGroup()
    model = kwargs.get('model')
    subset = kwargs.get('subset')
    chain = kwargs.get('chain')
    header = kwargs.get('header', False)
    get_bonds = kwargs.get('bonds',False) 
    altloc_sel = kwargs.get('altloc', None)
    
    assert isinstance(header, bool), 'header must be a boolean'

    if model is not None:
        if isinstance(model, int):
            if model < 0:
                raise ValueError('model must be greater than 0')
        else:
            raise TypeError('model must be an integer, {0} is invalid'
                            .format(str(model)))
    title_suffix = ''

    biomol = kwargs.get('biomol', False)
    hd = set_header(mmtf_struc)
    ag = set_info(ag, mmtf_struc, get_bonds, altloc_sel)

    if ag.numAtoms() > 0:
            LOGGER.report('{0} atoms and {1} coordinate set(s) were '
                          'parsed in %.2fs.'.format(ag.numAtoms(),
                                                    ag.numCoordsets()))
    else:
        ag = None
        LOGGER.warn('Atomic data could not be parsed, please '
                    'check the input file.')
        
    
        
    if ag is not None and isinstance(hd, dict):
        if biomol and _has_biomol(hd):
            ag = buildBiomolecules(hd, ag)
            if isinstance(ag, list):
                LOGGER.info('Biomolecular transformations were applied, {0} '
                            'biomolecule(s) are returned.'.format(len(ag)))
            else:
                LOGGER.info('Biomolecular transformations were applied to the '
                            'coordinate data.')    

    if header:
        return ag, hd
    else:
        return ag


def set_header(data_api):

    #get the transform here and convert it to the format that prody wants
    chain_list = data_api.chain_name_list
    assembly = _bio_transform(data_api)
    header = {
    'r_work': data_api.r_work,
    'r_free': data_api.r_free,
    'resolution': data_api.resolution,
    'title': data_api.title,
    'deposition_date': data_api.deposition_date,
    'release_date': data_api.release_date,
    'experimental_methods': data_api.experimental_methods,
    'biomoltrans': assembly
    }

    return header

def _bio_transform(dec):
    ret = {}
    matrix_line_format_string = ' %9.6f %9.6f %9.6f      %9.5f            \n'
    for t in dec.bio_assembly:
        L = []
        for trans in t['transformList']:
            chis = trans['chainIndexList']
            chains = sorted(set([dec.chain_name_list[c] for c in chis]))
            m = trans['matrix']
            L += [chains,
                  matrix_line_format_string % tuple(m[:4]),
                  matrix_line_format_string % tuple(m[4:8]),
                  matrix_line_format_string % tuple(m[8:12])]
        ret[t['name']] = L
    return ret

def set_info(atomgroup, mmtf_data, get_bonds=False, altloc_sel=None):

    mmtfHETATMtypes = set([
        "D-SACCHARIDE",
        "D-SACCHARIDE 1,4 AND 1,4 LINKING",
        "D-SACCHARIDE 1,4 AND 1,6 LINKING",
        "L-SACCHARIDE",
        "L-SACCHARIDE 1,4 AND 1,4 LINKING",
        "L-SACCHARIDE 1,4 AND 1,6 LINKING",
        "NON-POLYMER",
        "OTHER",
        "PEPTIDE-LIKE",
        "SACCHARIDE"])

    asize = mmtf_data.num_atoms

    if len(mmtf_data.chains_per_model) > 1:
        # get number of atoms in first model
        asize = 0
        groupIndex = 0
        modelChainCount = mmtf_data.chains_per_model[0]
        for chainGroupCount in mmtf_data.groups_per_chain[:modelChainCount]:
            #traverse groups
            for _ in range(chainGroupCount):
                group = mmtf_data.group_list[mmtf_data.group_type_list[groupIndex]]        
                asize += len(group['atomNameList'])
                groupIndex += 1

    fields = OrderedDict()

    x = mmtf_data.x_coord_list
    y = mmtf_data.y_coord_list
    z = mmtf_data.z_coord_list

    if len(x) != mmtf_data.num_models*asize:
        LOGGER.warn('Multi-model MMTF files with different molecules not supported.  Keeping only first model')
        coords = np.array([x, y, z]).T[:asize].reshape(1, asize, 3)
    else:     
        coords = np.array([x, y, z]).T.reshape(mmtf_data.num_models, asize, 3)

    # Initialize arrays for atom properties
    atom_names = np.empty(asize, dtype=ATOMIC_FIELDS['name'].dtype)
    resnames = np.empty(asize, dtype=ATOMIC_FIELDS['resname'].dtype)
    resnums = np.empty(asize, dtype=ATOMIC_FIELDS['resnum'].dtype)
    chainids = np.empty(asize, dtype=ATOMIC_FIELDS['chain'].dtype)
    segnames = np.empty(asize, dtype=ATOMIC_FIELDS['segment'].dtype)
    segnames[:] = ''
    elements = np.empty(asize, dtype=ATOMIC_FIELDS['element'].dtype)
    icodes = np.empty(asize, dtype=ATOMIC_FIELDS['icode'].dtype)

    hetero = np.zeros(asize, dtype=bool)
    termini = np.zeros(asize, dtype=bool)

    altlocs = np.array(mmtf_data.alt_loc_list[:asize], dtype=ATOMIC_FIELDS['altloc'].dtype)
    serials = np.array(mmtf_data.atom_id_list[:asize], dtype=ATOMIC_FIELDS['serial'].dtype)
    bfactors = np.array(mmtf_data.b_factor_list[:asize], dtype=ATOMIC_FIELDS['beta'].dtype)
    occupancies = np.array(mmtf_data.occupancy_list[:asize], dtype=ATOMIC_FIELDS['occupancy'].dtype)

    modelIndex = 0
    chainIndex = 0
    groupIndex = 0
    atomIndex = 0

    #traverse models
    for modelChainCount in mmtf_data.chains_per_model:
        chain_name_list = mmtf_data.chain_name_list

        #traverse chains
        for k in range(modelChainCount):
            chainGroupCount = mmtf_data.groups_per_chain[chainIndex]

            #traverse groups
            for _ in range(chainGroupCount):
                group = mmtf_data.group_list[mmtf_data.group_type_list[groupIndex]]
                groupAtomCount = len(group.get('atomNameList'))

                #traverse atoms
                for i in range(groupAtomCount):
                    atom_names[atomIndex] = group.get('atomNameList')[i]
                    elements[atomIndex] = group.get('elementList')[i]
                    resnames[atomIndex] = group.get('groupName')
                    resnums[atomIndex] = mmtf_data.group_id_list[groupIndex]
                    icodes[atomIndex] = mmtf_data.ins_code_list[groupIndex]
                    chainids[atomIndex] = chain_name_list[k]
                    if group.get('chemCompType') in mmtfHETATMtypes:
                        hetero[atomIndex] = True

                    atomIndex += 1

                groupIndex += 1
            chainIndex += 1
        modelIndex += 1
        break

    #detect termini based on chain changes
    termini[:-1] = chainids[1:] != chainids[:-1]
    termini[-1] = True  #the last atom is always a terminus

    mask = np.full(asize, True, dtype=bool)
    if altloc_sel != 'all':
        #mask out any unwanted alternative locations
        default_altloc = altloc_sel if altloc_sel != None else 'A'
        mask = (altlocs == '') | (altlocs == default_altloc)
        if np.all(mask == False) and altloc_sel == None and len(altlocs):
            #nothing selected, use first altloc; 6uwi
            mask = altlocs == altlocs[0]            
        
    atomgroup.setCoords(coords[:,mask])
    atomgroup.setNames(atom_names[mask])
    atomgroup.setResnums(resnums[mask])
    atomgroup.setResnames(resnames[mask])
    atomgroup.setChids(chainids[mask])
    atomgroup.setElements(elements[mask])
    atomgroup.setMasses(getMasses(elements[mask]))
    atomgroup.setBetas(bfactors[mask])
    atomgroup.setAltlocs(altlocs[mask])
    atomgroup.setOccupancies(occupancies[mask])
    atomgroup.setFlags('hetatm', hetero[mask])
    atomgroup.setFlags('pdbter', termini[mask])
    atomgroup.setFlags('selpdbter', termini[mask])
    atomgroup.setSerials(serials[mask])
    atomgroup.setIcodes(icodes[mask])
    atomgroup.setSegnames(segnames[mask])
    atomgroup.setTitle(mmtf_data.structure_id)
    
    if get_bonds and hasattr(mmtf_data,'bond_atom_list'):
        #have to remap any masked out atoms
        remaining = np.arange(asize)[mask]
        remap = np.full(asize,-1)
        remap[remaining] = np.arange(len(remaining))
        allbonds = np.array(mmtf_data.bond_atom_list).reshape(-1,2)
        nonpeptide = []
        #irgnore bonds between C and N in adjacent residues
        for a,b in allbonds:
            if a < asize and b < asize and mask[a] and mask[b] :
                if atom_names[a] != 'N' or atom_names[b] != 'C' or resnums[a]-resnums[b] != 1:
                    nonpeptide.append((remap[a],remap[b]))
        atomgroup.setBonds(nonpeptide)

    return atomgroup

def writeMMTF(filename, atoms, csets=None, autoext=True, **kwargs):
    """Write *atoms* in MMTF format to a file with name *filename* and return
    *filename*.  If *filename* ends with :file:`.gz`, a compressed file will
    be written.
    
    :arg atoms: an object with atom and coordinate data
    :type atoms: :class:`.Atomic`

    :arg csets: coordinate set indices, default is all coordinate sets

    :arg autoext: when not present, append extension :file:`.mmtf` to *filename*

    :keyword header: header to write too
    :type header: dict
    """
    try:
        from Bio.PDB.mmtf import MMTFIO
    except ImportError:
        raise ImportError('Biopython MMTFIO could not be imported. '
            'Reinstall ProDy or install Biopython and mmtf-python'
            'to solve the problem.')

    header = kwargs.get('header', None)

    if autoext and not filename.lower().endswith('.mmtf'):
        filename += '.mmtf'

    structure = atoms.toBioPythonStructure(header=header, csets=csets)
    io=MMTFIO()
    io.set_structure(structure)
    io.save(filename)
    return filename



"""This module defines miscellaneous utility functions that is public to users."""

import numpy as np

from prody import PY3K
from .misctools import addEnds, interpY, index, isListLike
from .checkers import checkCoords
from .logger import LOGGER


__all__ = ['calcTree', 'clusterMatrix', 
           'showLines', 'showMatrix', 'showBars', 
           'reorderMatrix', 'findSubgroups', 'getCoords',  
           'getLinkage', 'getTreeFromLinkage', 'clusterSubfamilies', 
           'calcRMSDclusters', 'calcGromosClusters', 'calcGromacsClusters', 
           'printAtomicMatrix', 'calcKmedoidClusters']

class LinkageError(Exception):
    pass

def clusterSubfamilies(similarities, n_clusters=0, linkage='all', method='tsne', cutoff=0.0, **kwargs):
    """Perform clustering based on members of the *ensemble* projected into lower a reduced
    dimension.
    
    :arg similarities: a matrix of similarities for each structure in the ensemble, such as
                        RMSD-matrix, dynamics-based spectral overlap, sequence similarity
    :type similarities: :class:`~numpy.ndarray`

    :arg n_clusters: the number of clusters to generate. If **0**, will scan a range of 
                        number of clusters and return the best one based on highest
                        silhouette score. Default is **0**.
    :type n_clusters: int

    :arg linkage: if **all**, will test all linkage types (ward, average, complete,
                    single). Otherwise will use only the one(s) given as input. Default is
                    **all**.
    :type linkage: str, list, tuple, :class:`~numpy.ndarray`

    :arg method: if set to **spectral**, will generate a Kirchoff matrix based on the 
                    cutoff value given and use that as input as clustering instead of
                    the values themselves. Default is **tsne**.
    :type method: str

    :arg cutoff: only used if *method* is set to **spectral**. This value is used for 
                    generating the Kirchoff matrix to use for generating clusters when
                    doing spectral clustering. Default is **0.0**.
    :type cutoff: float
    """

    # Import necessary packages
    try:
        from sklearn.manifold import SpectralEmbedding
        from sklearn.cluster import AgglomerativeClustering
        from sklearn.metrics import silhouette_score
        from sklearn.manifold import TSNE
    except ImportError:
        raise ImportError('need sklearn module')
        

    # Check inputs to make sure are of valid types/values
    if not isinstance(similarities, np.ndarray):
        raise TypeError('similarities should be a numpy ndarray')

    dim = similarities.shape
    if dim[0] != dim[1]:
        raise ValueError('similarities must be a square matrix')

    if n_clusters != 0:
        if not isinstance(n_clusters, int):
            raise TypeError('clusters must be an instance of int')
        if n_clusters < 1:
            raise ValueError('clusters must be a positive integer')
        elif n_clusters > similarities.shape[0]:
            raise ValueError('clusters can\'t be longer than similarities matrix')
        nclusts = range(n_clusters,n_clusters+1)
    else:
        nclusts = range(2,10,1)

    if linkage != 'all':
        # Check if given input for linkage is list-like
        if isListLike(linkage):
            for val in linkage:
                if val.lower() not in ['ward', 'average', 'complete', 'single']:
                    raise ValueError('linkage must be one or more of: \'ward\', \'average\', \'complete\', or \'single\'')
            if len(linkage) > 4:
                raise ValueError('linkage must be one or more of: \'ward\', \'average\', \'complete\', or \'single\'')
            linkages = [ x.lower() for x in linkage ]

        # If not, check if it is a valid string and method name
        else:
            if not isinstance(linkage, str):
                raise TypeError('linkage must be an instance of str or list-like of strs')

            if linkage not in ['ward', 'average', 'complete', 'single']:
                raise ValueError('linkage must one or more of: \'ward\', \'average\', \'complete\', or \'single\'')

            linkages = [linkage]
    else:
        linkages = ['ward', 'average', 'complete', 'single']

    if method != 'tsne':
        if not isinstance(method, str):
            raise TypeError('method must be an instance of str')
        if method != 'spectral':
            raise ValueError('method must be either \'tsne\' or \'spectral\'')

        if not isinstance(cutoff, float):
            raise TypeError('cutoff must be an instance of float')

    best_score = -1
    best_nclust = 0
    best_link = ''
    best_labels = []

    # Scan over range of clusters
    for x in nclusts:
        if method == 'tsne':
            embedding = TSNE(n_components=2)
            transform = embedding.fit_transform(similarities)

        else:
            kirchhoff = np.where(similarities > cutoff, 0, -1)
            embedding = SpectralEmbedding(n_components=2)
            transform = embedding.fit_transform(kirchhoff)

        for link in linkages:
            clustering = AgglomerativeClustering(linkage=link, n_clusters=x)
            clustering.fit(transform)

            silhouette_avg = silhouette_score(transform, clustering.labels_)
            
            if silhouette_avg > best_score:
                best_score = silhouette_avg
                best_nclust = x
                best_link = link
                best_labels = clustering.labels_


    return best_labels

def getCoords(data):
    """Get coordinates from *data* if possible and handle errors well.

    :arg data: a coordinate set or an object with ``getCoords`` method
    :type data: :class:`numpy.ndarray`, :class:`Atomic`, 
        :class:`Ensemble`, :class:`Trajectory`
    """
    try:
        data = (data._getCoords() if hasattr(data, '_getCoords') else
                data.getCoords())
    except AttributeError:
        try:
            checkCoords(data)
        except TypeError:
            raise TypeError('data must be a Numpy array or an object '
                            'with `getCoords` method')

    return data

def getLinkage(names, tree):
    """ Obtain the :func:`~scipy.cluster.hierarchy.linkage` matrix encoding 
    ``tree``. 
    
    :arg names: a list of names, the order determines the values in the 
                linkage matrix
    :type names: list, :class:`~numpy.ndarray`

    :arg tree: tree to be converted
    :type tree: :class:`~Bio.Phylo.BaseTree.Tree`
    """

    tree_terminals = tree.get_terminals()

    if len(tree_terminals) != len(names):
        raise ValueError('inconsistent number of terminals in tree and names')
    
    terminals = [None] * len(names)
    for clade in tree_terminals:
        i = index(names, clade.name)
        terminals[i] = clade

    n = len(terminals)
    nonterminals = [c for c in reversed(tree.get_nonterminals())]
    if len(nonterminals) != n-1:
        raise LinkageError('wrong number of terminal clades')

    Z = np.zeros((n-1, 4))

    root = tree.root

    def _indexOfClade(clade):
        if clade.is_terminal():
            i = index(terminals, clade)
        else:
            i = index(nonterminals, clade) + n
        return i

    def _height_of(clade):
        if clade.is_terminal():
            height = 0 
        else:
            height = max(_height_of(c) + c.branch_length for c in clade.clades)

        return height

    def _dfs(clade):
        if clade.is_terminal():
            return

        i = _indexOfClade(clade)
        clade_a = clade.clades[0]
        clade_b = clade.clades[1]

        a = _indexOfClade(clade_a)
        b = _indexOfClade(clade_b) 

        l = min(a, b)
        r = max(a, b)

        Z[i-n, 0] = l
        Z[i-n, 1] = r
        Z[i-n, 2] = _height_of(clade) * 2.
        Z[i-n, 3] = clade.count_terminals()

        _dfs(clade_a)
        _dfs(clade_b)
    
    _dfs(root)

    return Z

def getTreeFromLinkage(names, linkage):
    """ Obtain the tree encoded by ``linkage``. 
    
    :arg names: a list of names, the order should correspond to the values in  
                linkage
    :type names: list, :class:`~numpy.ndarray`

    :arg linkage: linkage matrix
    :type linkage: :class:`~numpy.ndarray`
    """
    try: 
        from Bio.Phylo.BaseTree import Tree, Clade
    except ImportError:
        raise ImportError('Phylo module could not be imported. '
            'Reinstall ProDy or install Biopython '
            'to solve the problem.')
    
    if not isinstance(linkage, np.ndarray):
        raise TypeError('linkage must be a numpy.ndarray instance')

    if linkage.ndim != 2:
        raise LinkageError('linkage must be a 2-dimensional matrix')

    if linkage.shape[1] != 4:
        raise LinkageError('linkage must have exactly 4 columns')

    n_terms = len(names)
    if linkage.shape[0] != n_terms-1:
        raise LinkageError('linkage must have exactly len(names)-1 rows')
    
    clades = []
    heights = []
    for name in names:
        clade = Clade(None, name)
        clades.append(clade)
        heights.append(0.)

    for link in linkage:
        l = int(link[0])
        r = int(link[1])
        height = link[2]

        left = clades[l]
        right = clades[r]

        lh = heights[l]
        rh = heights[r]

        left.branch_length = height - lh
        right.branch_length = height - rh

        clade = Clade(None, None)
        clade.clades.append(left)
        clade.clades.append(right)

        clades.append(clade)
        heights.append(height)

    return Tree(clade)

def calcTree(names, distance_matrix, method='upgma', linkage=False):
    """ Given a distance matrix, it creates an returns a tree structure.

    :arg names: a list of names
    :type names: list, :class:`~numpy.ndarray`

    :arg distance_matrix: a square matrix with length of ensemble. If numbers does not match *names*
                          it will raise an error
    :type distance_matrix: :class:`~numpy.ndarray`

    :arg method: method used for constructing the tree. Acceptable options are ``"upgma"``, ``"nj"``, 
                 or methods supported by :func:`~scipy.cluster.hierarchy.linkage` such as ``"single"``, 
                 ``"average"``, ``"ward"``, etc. Default is ``"upgma"``
    :type method: str

    :arg linkage: whether the linkage matrix is returned. Note that NJ trees do not support linkage
    :type linkage: bool
    """
            
    from .TreeConstruction import DistanceMatrix, DistanceTreeConstructor
    
    if len(names) != distance_matrix.shape[0] or len(names) != distance_matrix.shape[1]:
        raise ValueError("Mismatch between the sizes of matrix and names.")
    
    method = method.lower().strip()

    if method in ['ward', 'single', 'average', 'weighted', 'centroid', 'median', 'complete']:
        from scipy.cluster.hierarchy import linkage as hlinkage
        from scipy.spatial.distance import squareform
        
        Z = hlinkage(squareform(distance_matrix), method=method)
        tree = getTreeFromLinkage(names, Z)
    else:
        matrix = []
        k = 1
        Z = None
        for row in distance_matrix:
            matrix.append(list(row[:k]))
            k = k + 1
        
        if isinstance(names, np.ndarray):
            names = names.tolist()
        dm = DistanceMatrix(names, matrix)
        constructor = DistanceTreeConstructor()

        method = method.strip().lower()
        if method == 'nj':
            tree = constructor.nj(dm)
        elif method == 'upgma':
            tree = constructor.upgma(dm)
            if linkage:
                Z = getLinkage(names, tree)
        else:
            raise ValueError('Method can be only either "nj", "upgma" or '
                             'hierarchical clustering such as "single", "average", etc.')

        for node in tree.get_nonterminals():
            node.name = None

    if linkage:
        return tree, Z
    else:
        return tree

def writeTree(filename, tree, format_str='newick'):
    """ Write a tree to file using Biopython.

    :arg filename: name for output file
    :type filename: str

    :arg tree: a square matrix with length of ensemble. If numbers does not match *names*
                          it will raise an error
    :type tree: :class:`~Bio.Phylo.BaseTree.Tree`

    :arg format_str: a string specifying the format for the tree
    :type format_str: str
    """
    try: 
        from Bio import Phylo
    except ImportError:
        raise ImportError('Phylo module could not be imported. '
            'Reinstall ProDy or install Biopython '
            'to solve the problem.')

    if not isinstance(filename, str):
        raise TypeError('filename should be a string')

    if not isinstance(tree, Phylo.BaseTree.Tree):
        raise TypeError('tree should be a Biopython.Phylo Tree object')

    if not isinstance(format_str, str):
        raise TypeError('format_str should be a string')

    Phylo.write(tree, filename, format_str)


def clusterMatrix(distance_matrix=None, similarity_matrix=None, labels=None, return_linkage=None, **kwargs):
    """
    Cluster a distance matrix using scipy.cluster.hierarchy and 
    return the sorted matrix, indices used for sorting, sorted labels (if **labels** are passed),  
    and linkage matrix (if **return_linkage** is **True**). 
    
    :arg distance_matrix: an N-by-N matrix containing some measure of distance 
         such as 1. - seqid_matrix (Hamming distance), rmsds, or distances in PCA space
    :type distance_matrix: :class:`~numpy.ndarray`

    :arg similarity_matrix: an N-by-N matrix containing some measure of similarity 
         such as sequence identity, mode-mode overlap, or spectral overlap.
         Each element will be subtracted from 1. to get distance, so make sure this 
         is reasonable.
    :type similarity_matrix: :class:`~numpy.ndarray`
    
    :arg labels: labels for each matrix row that can be returned sorted
    :type labels: list

    :arg no_plot: if **True**, don't plot the dendrogram.
         default is **True**
    :type no_plot: bool
    
    :arg reversed: if set to **True**, then the sorting indices will be reversed.
    :type reversed: bool

    Other arguments for :func:`~scipy.hierarchy.linkage` and :func:`~scipy.hierarchy.dendrogram`
    can also be provided and will be taken as **kwargs**.
    """

    import scipy.cluster.hierarchy as sch
    from scipy import spatial
    if similarity_matrix is None and distance_matrix is None:
        raise ValueError('Please provide a distance matrix or a similarity matrix')
    
    orientation = kwargs.pop('orientiation', 'right')
    reversed = kwargs.pop('reversed', False)
    no_plot = kwargs.pop('no_plot', True)

    if distance_matrix is None:
        matrix = similarity_matrix
        distance_matrix = 1. - similarity_matrix
    else:
        matrix = distance_matrix
        
    formatted_distance_matrix = spatial.distance.squareform(distance_matrix)
    linkage_matrix = sch.linkage(formatted_distance_matrix, **kwargs)
    sorting_dendrogram = sch.dendrogram(linkage_matrix, orientation=orientation, labels=labels, no_plot=no_plot)

    indices = sorting_dendrogram['leaves']
    sorted_labels = sorting_dendrogram['ivl']

    if reversed:
        indices = indices[::-1]
        sorted_labels = sorted_labels[::-1]
    
    sorted_matrix = matrix[indices, :]
    sorted_matrix = sorted_matrix[:, indices]
    
    return_vals = [sorted_matrix, indices]

    if labels is not None:
        return_vals.append(sorted_labels)
    if return_linkage:
        return_vals.append(linkage_matrix)
    return tuple(return_vals) # convert to tuple to avoid [pylint] E0632:Possible unbalanced tuple unpacking

def showLines(*args, **kwargs):
    """
    Show 1-D data using :func:`~matplotlib.axes.Axes.plot`. 
    
    :arg x: (optional) x coordinates. *x* can be an 1-D array or a 2-D matrix of 
            column vectors.
    :type x: :class:`~numpy.ndarray`

    :arg y: data array. *y* can be an 1-D array or a 2-D matrix of 
            column vectors.
    :type y: :class:`~numpy.ndarray`

    :arg dy: an array of variances of *y* which will be plotted as a 
             band along *y*. It should have the same shape with *y*.
    :type dy: :class:`~numpy.ndarray`

    :arg lower: an array of lower bounds which will be plotted as a 
                band along *y*. It should have the same shape with *y* and should be 
                paired with *upper*.
    :type lower: :class:`~numpy.ndarray`

    :arg upper: an array of upper bounds which will be plotted as a 
                band along *y*. It should have the same shape with *y* and should be 
                paired with *lower*.
    :type upper: :class:`~numpy.ndarray`

    :arg alpha: the transparency of the band(s) for plotting *dy*.
    :type alpha: float

    :arg beta: the transparency of the band(s) for plotting *miny* and *maxy*.
    :type beta: float

    :arg ticklabels: user-defined tick labels for x-axis.
    :type ticklabels: list
    """
    
    # note for developers: this function serves as a low-level 
    # plotting function which provides basic utilities for other 
    # plotting functions. Therefore showFigure is not handled 
    # in this function as it should be already handled in the caller.

    ticklabels = kwargs.pop('ticklabels', None)
    dy = kwargs.pop('dy', None)
    miny = kwargs.pop('lower', None)
    maxy = kwargs.pop('upper', None)
    alpha = kwargs.pop('alpha', 0.5)
    beta = kwargs.pop('beta', 0.25)
    gap = kwargs.pop('gap', False)
    labels = kwargs.pop('label', None)

    from matplotlib import cm, ticker
    from matplotlib.pyplot import figure, gca, xlim
    from .drawtools import IndexFormatter

    ax = gca()
    lines = ax.plot(*args, **kwargs)

    polys = []
        
    for i, line in enumerate(lines):
        color = line.get_color()
        x, y = line.get_data()
        
        if gap:
            x_new, y_new = addEnds(x, y)
            line.set_data(x_new, y_new)
        else:
            x_new, y_new = x, y
        
        if labels is not None:
            if np.isscalar(labels):
                line.set_label(labels)
            else:
                try:
                    line.set_label(labels[i])
                except IndexError:
                    raise ValueError('The number of labels ({0}) and that of y ({1}) do not match.'
                                     .format(len(labels), len(line)))
        
        # the following function needs to be here so that line exists
        def sub_array(a, i, tag='a'):
            ndim = 0
            if a is not None:
                if np.isscalar(a[0]):
                    ndim = 1   # a plain list (array)
                else:
                    ndim = 2   # a nested list (array)
            else:
                return None

            if ndim == 1:
                _a = a
            else:
                try:
                    _a = a[i]
                except IndexError:
                    raise ValueError('The number of {2} ({0}) and that of y ({1}) do not match.'
                                     .format(len(miny), len(line), tag))

            if len(_a) != len(y):
                raise ValueError('The shapes of {2} ({0}) and y ({1}) do not match.'
                                 .format(len(_miny), len(y), tag))
            return _a

        if miny is not None and maxy is not None:
            _miny = sub_array(miny, i)
            _maxy = sub_array(maxy, i)

            if gap:
                _, _miny = addEnds(x, _miny)
                _, _maxy = addEnds(x, _maxy)
                
            poly = ax.fill_between(x_new, _miny, _maxy,
                                    alpha=beta, facecolor=color, edgecolor=None,
                                    linewidth=1, antialiased=True)
            polys.append(poly)

        if dy is not None:
            _dy = sub_array(dy, i)

            if gap:
                _, _dy = addEnds(x, _dy)
                
            poly = ax.fill_between(x_new, y_new-_dy, y_new+_dy,
                                    alpha=alpha, facecolor=color, edgecolor=None,
                                    linewidth=1, antialiased=True)
            polys.append(poly)

    ax.margins(x=0)
    if ticklabels is not None:
        if callable(ticklabels):
            ax.get_xaxis().set_major_formatter(ticker.FuncFormatter(ticklabels))
        else:
            ax.get_xaxis().set_major_formatter(IndexFormatter(ticklabels))
    
    ax.xaxis.set_major_locator(ticker.AutoLocator())
    ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())

    return lines, polys

def showBars(ydata, xdata=None, *args, **kwargs):
    """
    Show 1-D data using :func:`~matplotlib.axes.Axes.bar`. 
    
    :arg x: (optional) x coordinates. *x* can be an 1-D array or a 2-D matrix of 
            column vectors.
            If not provided, a range with the length of the y data will be used.
    :type x: :class:`~numpy.ndarray`

    :arg y: data array. *y* can be an 1-D array or a 2-D matrix of 
            column vectors.
    :type y: :class:`~numpy.ndarray`

    :arg ticklabels: user-defined tick labels for x-axis.
    :type ticklabels: list
    """
    
    # note for developers: this function serves as a low-level 
    # plotting function which provides basic utilities for other 
    # plotting functions. Therefore showFigure is not handled 
    # in this function as it should be already handled in the caller.

    ticklabels = kwargs.pop('ticklabels', None)
    gap = kwargs.pop('gap', False)
    labels = kwargs.pop('label', None)

    from matplotlib import cm, ticker
    from matplotlib.pyplot import figure, gca, xlim
    from .drawtools import IndexFormatter

    if xdata is None:
        xdata = list(range(len(ydata)))

    ax = gca()
    bars = ax.bar(xdata, ydata, *args, **kwargs)
        
    for i, bar in enumerate(bars):
        x, y = bar.get_xy()
        
        if gap:
            x_new, y_new = addEnds(x, y)
            bar.set_data(x_new, y_new)
        else:
            x_new, y_new = x, y
        
        if labels is not None:
            if np.isscalar(labels):
                bar.set_label(labels)
            else:
                try:
                    bar.set_label(labels[i])
                except IndexError:
                    raise ValueError('The number of labels ({0}) and that of y ({1}) do not match.'
                                     .format(len(labels), len(bar)))

    ax.margins(x=0)
    if ticklabels is not None:
        if callable(ticklabels):
            ax.get_xaxis().set_major_formatter(ticker.FuncFormatter(ticklabels))
        else:
            ax.get_xaxis().set_major_formatter(IndexFormatter(ticklabels))
    
    ax.xaxis.set_major_locator(ticker.AutoLocator())
    ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())

    return bars

def showMatrix(matrix, x_array=None, y_array=None, **kwargs):
    """Show a matrix using :meth:`~matplotlib.axes.Axes.imshow` or
    :meth:`~matplotlib.axes.Axes.scatter` if *markersize* is provided.

    Curves on x- and y-axis can be added.

    :arg matrix: matrix to be displayed
    :type matrix: :class:`~numpy.ndarray`

    :arg x_array: data to be plotted above the matrix
    :type x_array: :class:`~numpy.ndarray`

    :arg y_array: data to be plotted on the left side of the matrix
    :type y_array: :class:`~numpy.ndarray`

    :arg percentile: a percentile threshold to remove outliers, i.e. only showing data within *p*-th 
                     to *100-p*-th percentile
    :type percentile: float

    :arg vmin: a minimum value threshold to remove outliers, i.e. only showing data greater than vmin
               This overrides percentile.
    :type vmin: float

    :arg vmax: a maximum value threshold to remove outliers, i.e. only showing data less than vmax
               This overrides percentile.
    :type vmax: float

    :arg interactive: turn on or off the interactive options
    :type interactive: bool

    :arg xtickrotation: how much to rotate the xticklabels in degrees
                        default is 0
    :type xtickrotation: float

    :arg markersize: size of square markers for using :meth:`~matplotlib.axes.Axes.scatter`
        to help show matrices with small data regions compared to zeros.
        Note only non-zeros are plotted so the colorbar range may change if not using norm
        Default is None, which results in using :meth:`~matplotlib.axes.Axes.imshow`
    :type markersize: float
    """

    from matplotlib import ticker
    from matplotlib.gridspec import GridSpec
    from matplotlib.collections import LineCollection
    from matplotlib.pyplot import gca, sca, sci, colorbar, subplot

    from .drawtools import drawTree, IndexFormatter

    p = kwargs.pop('percentile', None)
    vmin = vmax = None
    if p is not None:
        vmin = np.percentile(matrix, p)
        vmax = np.percentile(matrix, 100-p)
    
    vmin = kwargs.pop('vmin', vmin)
    vmax = kwargs.pop('vmax', vmax)
    vcenter = kwargs.pop('vcenter', None)
    norm = kwargs.pop('norm', None)

    markersize = kwargs.pop('markersize', None)

    if vcenter is not None and norm is None:
        if PY3K:
            try:
                from matplotlib.colors import DivergingNorm
            except ImportError:
                from matplotlib.colors import TwoSlopeNorm as DivergingNorm

            norm = DivergingNorm(vmin=vmin, vcenter=vcenter, vmax=vmax)
        else:
            LOGGER.warn('vcenter cannot be used in Python 2 so norm remains None')

    lw = kwargs.pop('linewidth', 1)
    
    W = H = kwargs.pop('ratio', 6)

    ticklabels = kwargs.pop('ticklabels', None)
    xticklabels = kwargs.pop('xticklabels', ticklabels)
    yticklabels = kwargs.pop('yticklabels', ticklabels)

    xtickrotation = kwargs.pop('xtickrotation', 0.)

    show_colorbar = kwargs.pop('colorbar', True)
    cb_extend = kwargs.pop('cb_extend', 'neither')
    allticks = kwargs.pop('allticks', False) # this argument is temporary and will be replaced by better implementation
    interactive = kwargs.pop('interactive', True)

    import matplotlib
    if float(matplotlib.__version__.split('.')[0]) >= 3 or float(matplotlib.__version__.split('.')[1]) >= 6:
        LOGGER.warn('matplotlib 3.6 and later are not compatible with interactive matrices')
        interactive = False

    cmap = kwargs.pop('cmap', 'jet')
    origin = kwargs.pop('origin', 'lower')

    if not isinstance(x_array, np.ndarray) or not isinstance(y_array, np.ndarray):
        try:
            from Bio import Phylo
        except ImportError:
            raise ImportError('Phylo module could not be imported. '
                'Reinstall ProDy or install Biopython '
                'to solve the problem.')
        tree_mode_y = isinstance(y_array, Phylo.BaseTree.Tree)
        tree_mode_x = isinstance(x_array, Phylo.BaseTree.Tree)
    else:
        tree_mode_x = False
        tree_mode_y = False

    if x_array is not None and y_array is not None:
        nrow = 2; ncol = 2
        i = 1; j = 1
        width_ratios = [1, W]
        height_ratios = [1, H]
        aspect = 'auto'
    elif x_array is not None and y_array is None:
        nrow = 2; ncol = 1
        i = 1; j = 0
        width_ratios = [W]
        height_ratios = [1, H]
        aspect = 'auto'
    elif x_array is None and y_array is not None:
        nrow = 1; ncol = 2
        i = 0; j = 1
        width_ratios = [1, W]
        height_ratios = [H]
        aspect = 'auto'
    else:
        nrow = 1; ncol = 1
        i = 0; j = 0
        width_ratios = [W]
        height_ratios = [H]
        aspect = kwargs.pop('aspect', None)

    main_index = (i, j)
    upper_index = (i-1, j)
    left_index = (i, j-1)

    complex_layout = nrow > 1 or ncol > 1

    ax1 = ax2 = ax3 = None

    if complex_layout:
        gs = GridSpec(nrow, ncol, width_ratios=width_ratios, 
                      height_ratios=height_ratios, hspace=0., wspace=0.)

    ## draw matrix
    if complex_layout:
        ax3 = subplot(gs[main_index])
    else:
        ax3 = gca()
    
    if markersize is None:
        # default behaviour
        im = ax3.imshow(matrix, aspect=aspect, vmin=vmin, vmax=vmax,
                        norm=norm, cmap=cmap, origin=origin, **kwargs)
    else:
        zeros_matrix = np.zeros(matrix.shape)
        im = ax3.imshow(zeros_matrix, vmin=-1, vmax=1, cmap='seismic')
        plot_list = []
        for rows,cols in zip(np.where(matrix!=0)[0],np.where(matrix!=0)[1]):
            plot_list.append([cols,rows,matrix[rows,cols]])
        plot_list = np.array(plot_list)

        im = ax3.scatter(plot_list[:,0], plot_list[:,1], c=plot_list[:,2],
                         s=markersize, marker='s', cmap=cmap, norm=norm, **kwargs)

    if xticklabels is not None:
        ax3.xaxis.set_major_formatter(IndexFormatter(xticklabels))
    if yticklabels is not None and ncol == 1:
        ax3.yaxis.set_major_formatter(IndexFormatter(yticklabels))

    if allticks:
        ax3.xaxis.set_major_locator(ticker.IndexLocator(offset=0.5, base=1.))
        ax3.yaxis.set_major_locator(ticker.IndexLocator(offset=0.5, base=1.))
    else:
        locator = ticker.AutoLocator()
        locator.set_params(integer=True)
        minor_locator = ticker.AutoMinorLocator()

        ax3.xaxis.set_major_locator(locator)
        ax3.xaxis.set_minor_locator(minor_locator)

        locator = ticker.AutoLocator()
        locator.set_params(integer=True)
        minor_locator = ticker.AutoMinorLocator()
        
        ax3.yaxis.set_major_locator(locator)
        ax3.yaxis.set_minor_locator(minor_locator)

    if ncol > 1:
        ax3.yaxis.set_major_formatter(ticker.NullFormatter())
    
    ## draw x_ and y_array
    lines = []

    if nrow > 1:
        ax1 = subplot(gs[upper_index])

        if tree_mode_x:
            Y, X = drawTree(x_array, label_func=None, orientation='vertical', 
                            inverted=True)
            miny = min(Y.values())
            maxy = max(Y.values())

            minx = min(X.values())
            maxx = max(X.values())

            ax1.set_xlim(minx-.5, maxx+.5)
            ax1.set_ylim(miny, 1.05*maxy)
        else:
            ax1.set_xticklabels([])
            
            y = x_array
            xp, yp = interpY(y)
            points = np.array([xp, yp]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            lcy = LineCollection(segments, array=yp, linewidths=lw, cmap=cmap)
            lines.append(lcy)
            ax1.add_collection(lcy)

            ax1.set_xlim(xp.min()-.5, xp.max()+.5)
            ax1.set_ylim(yp.min(), yp.max())

        if ax3.xaxis_inverted():
            ax2.invert_xaxis()

        ax1.axis('off')

    if ncol > 1:
        ax2 = subplot(gs[left_index])
        
        if tree_mode_y:
            X, Y = drawTree(y_array, label_func=None, inverted=True)
            miny = min(Y.values())
            maxy = max(Y.values())

            minx = min(X.values())
            maxx = max(X.values())

            ax2.set_ylim(miny-.5, maxy+.5)
            ax2.set_xlim(minx, 1.05*maxx)
        else:
            ax2.set_xticklabels([])
            
            y = y_array
            xp, yp = interpY(y)
            points = np.array([yp, xp]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            lcx = LineCollection(segments, array=yp, linewidths=lw, cmap=cmap)
            lines.append(lcx)
            ax2.add_collection(lcx)
            ax2.set_xlim(yp.min(), yp.max())
            ax2.set_ylim(xp.min()-.5, xp.max()+.5)
        
        ax2.invert_xaxis()

        if ax3.yaxis_inverted():
            ax2.invert_yaxis()

        ax2.axis('off')

    ## draw colorbar
    sca(ax3)
    cb = None
    if show_colorbar:
        if nrow > 1:
            axes = [ax1, ax2, ax3]
            while None in axes:
                axes.remove(None)
            s = H / (H + 1.)
            cb = colorbar(mappable=im, ax=axes, anchor=(0, 0), shrink=s, extend=cb_extend)
        else:
            cb = colorbar(mappable=im, extend=cb_extend)

    sca(ax3)
    sci(im)

    if interactive:
        from prody.utilities import ImageCursor
        from matplotlib.pyplot import connect
        cursor = ImageCursor(ax3, im)
        connect('button_press_event', cursor.onClick)

    ax3.tick_params(axis='x', rotation=xtickrotation)

    return im, lines, cb

def reorderMatrix(names, matrix, tree, axis=None):
    """
    Reorder a matrix based on a tree and return the reordered matrix 
    and indices for reordering other things.

    :arg names: a list of names associated with the rows of the matrix
        These names must match the ones used to generate the tree
    :type names: list

    :arg matrix: any square matrix
    :type matrix: :class:`~numpy.ndarray`

    :arg tree: any tree from :func:`calcTree`
    :type tree: :class:`~Bio.Phylo.BaseTree.Tree`

    :arg axis: along which axis the matrix should be reordered. 
               Default is **None** which reorder along all the axes
    :type axis: int
    """

    try:
        from Bio import Phylo
    except ImportError:
        raise ImportError('Phylo module could not be imported. '
            'Reinstall ProDy or install Biopython '
            'to solve the problem.')

    try:
        if matrix.ndim != 2:
            raise ValueError('matrix should be a 2D matrix.')
    except AttributeError:
        raise TypeError('matrix should be a numpy array.')

    if np.shape(matrix)[0] != np.shape(matrix)[1]:
        raise ValueError('matrix should be a square matrix')
    
    names = np.asarray(names)

    if np.isscalar(names):
        raise TypeError('names should be list-like')
    
    if not len(names):
        raise TypeError('names is empty')

    if not isinstance(tree, Phylo.BaseTree.Tree):
        raise TypeError('tree should be a BioPython Tree')

    if len(names) != len(matrix):
        raise ValueError('names should have entries for each matrix row/column')
    
    terminals = tree.get_terminals()
    if len(names) != len(terminals):
        raise ValueError('names should have entries for each tree terminal')

    if len(terminals) != len(matrix):
        raise ValueError('matrix should have a row for each tree terminal')

    indices = []
    for terminal in terminals:
        name = terminal.name
        locs = np.where(names == name)[0]
        if not len(locs):
            raise ValueError('inconsistent names and tree: %s not in names'%name)

        if len(locs) > 1:
            raise ValueError('inconsistent names and tree: duplicate name %s in names'%name)
        indices.append(locs[0])

    if axis is not None:
        I = [np.arange(s) for s in matrix.shape] 
        axes = [axis] if np.isscalar(axis) else axis
        for ax in axes:
            I[ax] = indices
    else:
        I = [indices] * matrix.ndim
    
    rmatrix = matrix[np.ix_(*I)]
    
    return rmatrix, indices

def findSubgroups(tree, c, method='naive', **kwargs):
    """
    Divide **tree** into subgroups using a criterion **method** and a cutoff **c**.
    Returns a list of lists with labels divided into subgroups.
    """

    method = method.lower().strip()
    terminals = tree.get_terminals()
    names = [clade.name for clade in terminals]
    Z = None

    if method != 'naive':
        try:
            Z = getLinkage(names, tree)
        except LinkageError:
            print('Failed to build linkage; fall back to naive criterion')
            method = 'naive'
    
    if method == 'naive':
        subgroups = [[names[0]]]
        for i in range(len(terminals)-1):
            curr_clade = terminals[i]
            next_clade = terminals[i + 1]
            d = tree.distance(curr_clade, next_clade)
            if d > c:
                subgroups.append([])
            subgroups[-1].append(next_clade.name)
    else:
        from scipy.cluster.hierarchy import fcluster
        
        T = fcluster(Z, c, criterion=method, **kwargs)
        labels = np.unique(T)
        subgroups = [[] for _ in range(len(labels))]

        for i, t in enumerate(T):
            subgroups[t-1].append(names[i])

    return subgroups


def getAtomicTable(matrix, atoms_i=None, atoms_j=None, 
                   fmt='%5d', sep='\t'):
    """Generates a new table for a matrix with atom labels along 
    the top and at the beginning of each line for :func:`.printAtomicTable`.

    :arg matrix: any square 2D data with a size matching atoms
    :type matrix: tuple, list, :class:`~numpy.ndarray`

    :arg atoms_i: any :class:`.Atomic` object to label the rows
    :type atoms_i: :class:`.Atomic`

    :arg atoms_j: any :class:`.Atomic` object to label the columns
        uses atoms_i by default
    :type atoms_j: :class:`.Atomic`

    :arg fmt: format string for formatting numbers
    :type fmt: str
    """
    if not isListLike(matrix):
        raise TypeError('matrix should be list-like')

    matrix = np.array(matrix)
    if matrix.ndim != 2:
        raise ValueError('matrix should be 2-dimensional')

    if atoms_j is None:
        atoms_j = atoms_i

    if atoms_i is not None and matrix.shape[0] != atoms_i.numAtoms():
        raise ValueError('number of rows should be number of atoms_i')

    if atoms_j is not None and matrix.shape[1] != atoms_j.numAtoms():
        raise ValueError('number of cols should be number of atoms_j')

    if not isinstance(fmt, str):
        raise TypeError('fmt should be a string')

    chars = [list(item) for item in fmt.split('.')]
    nums = []
    for item in chars:
        num_str = ''
        for char in item:
            if char.isnumeric():
                num_str += char
        nums.append(int(num_str))

    if len(nums) == 1:
        length = nums[0]
    else:
        if nums[0] >= nums[1] + 2:
            length = nums[0]
        else:
            length = nums[1] + 2

    table = ' '*length

    if atoms_i is None and atoms_j is None:
        resnum_length = len(str(max(matrix.shape)))
        chid_length = 0

    elif atoms_i is None:
        resnum_length = len(str(max(atoms_j.getResnums())))
        chid_length = max([len(chid) for chid in atoms_j.getChids()])

    elif atoms_j is None:
        resnum_length = len(str(max(atoms_i.getResnums())))
        chid_length = max([len(chid) for chid in atoms_i.getChids()])

    else:        
        resnum_length = max(len(str(max(atoms_i.getResnums()))),
                            len(str(max(atoms_j.getResnums()))))
        chid_length = max(max([len(chid) for chid in atoms_i.getChids()]),
                        max([len(chid) for chid in atoms_j.getChids()]))
    
    for j in range(matrix.shape[1]):
        table += sep
        if length >= resnum_length + chid_length:
            if atoms_j is None:
                chid = ''
            else:
                chid = atoms_j[j].getChid()
            table += ' '*(chid_length - len(chid)) + chid
        if length >= resnum_length + chid_length + 1:
            table += ' '
        
        if atoms_j is not None:
            if length >= resnum_length + chid_length + 2:
                table += atoms_j[j].getResname()

            table += '%{0}d'.format(resnum_length) % atoms_j[j].getResnum()
    table += '\n'

    for i, row in enumerate(matrix):
        if atoms_i is not None:
            table += '\t{} {}'.format(atoms_i[i].getChid(),
                                      atoms_i[i].getResname())
            table += '%{0}d'.format(resnum_length) % atoms_i[i].getResnum()
        for element in row:
            table += fmt % element
        table += '\n'

    return table


def printAtomicMatrix(matrix, atoms=None, step=10, 
                      fmt='%8d', sep='\t'):
    """Prints a new table for a matrix with
    atom labels along the top and at the 
    beginning of each line.

    :arg matrix: any square 2D data with a size matching atoms
    :type matrix: tuple, list, :class:`~numpy.ndarray`

    :arg atoms: any :class:`.Atomic` object to label the data
    :type atoms: :class:`.Atomic`
    """
    attempts = len(matrix)//step
    if len(matrix) > step * attempts:
        attempts += 1
    for i in range(attempts):
        start = step * i
        stop = step * (i+1)
        submatrix = matrix[:,start:stop,]
        if atoms is not None:
            atoms_i = atoms
            atoms_j = atoms[start:stop]
        else:
            atoms_i = atoms
            atoms_j = atoms
        print(getAtomicTable(submatrix, atoms_i, atoms_j,
                             fmt, sep))

    return


def calcRMSDclusters(rmsd_matrix, c, labels=None):
    """
    Divide **rmsd_matrix** into clusters using the gromos method 
    with a cutoff **c** as implemented in gromacs (see 
    https://manual.gromacs.org/documentation/current/onlinehelp/gmx-cluster.html)

    Returns a list of lists with labels divided into clusters.
    """    
    clusters = []

    useful_rmsd_matrix = rmsd_matrix
    indices = list(range(len(rmsd_matrix)))
    if labels is None:
        elements = indices
    else:
        elements = labels

    while len(elements) > 0:
        neighbours = []
        num_neighbours = np.zeros(len(elements))
        for i, elem in enumerate(elements):
            neighbours_i = list(np.array(elements)[list(np.where(useful_rmsd_matrix[i] <= c)[0])])
            neighbours_i.pop(neighbours_i.index(elem))
            neighbours.append(neighbours_i)
            num_neighbours[i] = len(neighbours_i)

        argmax_num_n = np.argmax(num_neighbours)
        argmax_elem = elements[argmax_num_n]

        clusters.append([])
        for i, elem in enumerate(elements):
            if argmax_elem in neighbours[i] or elem == argmax_elem:
                clusters[-1].append(elem)

                elements = list(elements)
                indices.pop(elements.index(elem))
                elements.pop(elements.index(elem))
                elements = np.array(elements)

        useful_rmsd_matrix = rmsd_matrix[:, list(indices)][list(indices), :]

    return clusters

calcGromosClusters = calcRMSDclusters
calcGromacsClusters = calcRMSDclusters

def calcKmedoidClusters(coordsets, nClusters):
    try:
        from sklearn_extra.cluster import KMedoids
    except ImportError:
        raise ImportError('Please install scikit-learn-extra to run this function')
    
    X = coordsets.reshape(coordsets.shape[0], -1)
    c = KMedoids(n_clusters=nClusters, random_state=0).fit(X)
    labels = c.labels_
    _, counts = np.unique(labels, return_counts=True)
    return c.medoid_indices_, labels, counts
